#include <stddef.h>
#include <inttypes.h>
#include <avr\io.h>
#include <avr\wdt.h>
#include <avr\signal.h>                            
#include <avr\pgmspace.h>
#include "uart.h"


#ifdef UART0_ISR
static uint8_t uart0_rx_buffer[UART0_RX_BUFFER_SIZE];
volatile static uint8_t uart0_rx_head, uart0_rx_tail;
static uint8_t uart0_tx_buffer[UART0_TX_BUFFER_SIZE];
volatile static uint8_t uart0_tx_head, uart0_tx_tail;
#endif

#ifdef UART1_ISR
static uint8_t uart1_rx_buffer[UART1_RX_BUFFER_SIZE];
volatile static uint8_t uart1_rx_head, uart1_rx_tail;
static uint8_t uart1_tx_buffer[UART1_TX_BUFFER_SIZE];
volatile static uint8_t uart1_tx_head, uart1_tx_tail;
#endif

#if !defined(UCSR0A)
  #if defined(UCSRA)
    #define UCSR0A  UCSRA
  #elif defined(USR)
    #define UCSR0A  USR
  #endif
#endif


#if !defined(UCSR0B)
  #if defined(UCSRB)
    #define UCSR0B  UCSRB
  #elif defined(UCR)
    #define UCSR0B  UCR
  #endif
#endif

#if !defined(UCSR0C) & defined(UCSRC)
  #define UCSR0C  UCSRC
#endif

#if !defined(UBRR0L)
  #if defined(UBRRL)
    #define UBRR0L  UBRRL
  #elif defined(UBRR)
    #define UBRR0L  UBRR
  #elif defined(UBRR0)
    #define UBRR0L  UBRR0
  #endif
#endif

#if !defined(UBRR0H) & defined(UBRRH)
  #define UBRR0H  UBRRH
#endif

#if !defined(UCSZ00) & defined(UCSZ0)
  #define UCSZ00  UCSZ0
#endif

#if !defined(UCSZ01) & defined(UCSZ1)
  #define UCSZ01  UCSZ1
#endif

#if !defined(TXEN0) & defined(TXEN)
  #define TXEN0  TXEN
#endif

#if !defined(RXEN0) & defined(RXEN)
  #define RXEN0  RXEN
#endif

#if !defined(RXCIE0) & defined(RXCIE)
  #define RXCIE0  RXCIE
#endif

#if !defined(UDRIE0) & defined(UDRIE)
  #define UDRIE0  UDRIE
#endif

#if !defined(FE0) & defined(FE)
  #define FE0  FE
#endif

#if !defined(DOR0) & defined(DOR)
  #define DOR0  DOR
#endif

#if !defined(UDR0) & defined(UDR)
  #define UDR0  UDR
#endif

#if !defined(UDRE0) & defined(UDRE)
  #define UDRE0  UDRE
#endif

#if !defined(RXC0) & defined(RXC)
  #define RXC0  RXC
#endif

#if !defined(UBRR1L) & defined(UBRR1)
  #define UBRR1L  UBRR1
#endif

#if !defined(TXEN1) & defined(TXEN)
  #define TXEN1 TXEN
#endif

#if !defined(RXEN1) & defined(RXEN)
  #define RXEN1 RXEN
#endif

#if !defined(RXCIE1) & defined(RXCIE)
  #define RXCIE1 RXCIE
#endif

#if !defined(UDRIE1) & defined(UDRIE)
  #define UDRIE1 UDRIE
#endif

#if !defined(FE1) & defined(FE)
  #define FE1 FE
#endif

#if !defined(DOR1) & defined(DOR)
  #define DOR1 DOR
#endif

#if !defined(UDRE1) & defined(UDRE)
  #define UDRE1 UDRE
#endif

#if !defined(RXC1) & defined(RXC)
  #define RXC1 RXC
#endif


#ifdef UART0
void uart0_Init(uint16_t baud) {

// todo: parity, bits
  UCSR0B = 0;
  UCSR0A = 0;
#if defined(U2X0)
  if (baud & (1 << 15)) UCSR0A = (1 << U2X0);
#endif
  UBRR0L = baud & 0xFF;
#if defined(UBRR0H)
  UBRR0H = (baud >> 8) & 0x0F;
#endif
#if defined(UCSR0C)
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00) | (1 << UPM01) | (1 << UPM00);
#endif
#ifdef UART0_ISR
  uart0_tx_head = 0;
  uart0_tx_tail = 0;
  uart0_rx_head = 0;
  uart0_rx_tail = 0;
  UCSR0B = (1 << TXEN0) | (1 << RXEN0) | (1 << RXCIE0);
#else
  UCSR0B = (1 << TXEN0) | (1 << RXEN0);
#endif
}

int16_t uart0_Put(uint8_t value) {

#ifdef UART0_ISR
  uint8_t h = uart0_tx_head;
  uint8_t i = h + 1;
  i &= UART0_TX_BUFFER_SIZE -1;
  while (1) {
    if (i != uart0_tx_tail) break;
    wdt_reset();
  }
  uart0_tx_buffer[h] = value;
  uart0_tx_head = i;
  UCSR0B |= (1 << UDRIE0);
#else
  while (1) {
    if (UCSR0A & (1 << UDRE0)) break;
    wdt_reset();
  }
  UDR0 = value;
#endif
  return value;
}

void uart0_Write(const int8_t *Data, int16_t Count) {

   while (1) {
     if (Count <= 0) break;
     uart0_Put(*Data++);
     Count--;
   }
}

int16_t uart0_TxSpace(void) {

#ifdef UART0_ISR
  int16_t i = uart0_tx_tail - uart0_tx_head;
  if (i <= 0) i += UART0_TX_BUFFER_SIZE;
  return i - 1;
#else
  return 255;
#endif
}

void uart0_TxFlush(void) {

#ifdef UART0_ISR
  UCSR0B &= ~(1 << UDRIE0);
  uart0_tx_head = 0;
  uart0_tx_tail = 0;
  UCSR0B |= (1 << UDRIE0);
#endif
}

void uart0_RxClear(void) {

#ifdef UART0_ISR
  UCSR0B &= ~(1 << RXCIE0);
  uart0_rx_head = 0;
  uart0_rx_tail = 0;
// todo: error handling reseting
  UCSR0B |= (1 << RXCIE0);
#endif
}

int16_t uart0_RxCount(void) {

#ifdef UART0_ISR
  int16_t i = uart0_rx_head - uart0_rx_tail;
  if (i < 0) i += UART0_RX_BUFFER_SIZE;
  return i;
#else
  return ((UCSR0A & (1 << RXC0)) ? 1 : 0);
#endif
}

int16_t uart0_Get(void) {

#ifdef UART0_ISR
  uint8_t i = uart0_rx_tail;
  while (1) {
    if (i != uart0_rx_head) break;
    wdt_reset();
  }
  uint8_t r = uart0_rx_buffer[i++];
  i &= UART0_RX_BUFFER_SIZE -1;
  uart0_rx_tail = i;
  UCSR0B |= (1 << RXCIE0);
  return r;
#else
  while (1) {
    if (UCSR0A & (1 << RXC0)) break;
    wdt_reset();
  }
  uint8_t r = UDR0;
  if (UCSR0A & ((1 << FE0) | (1 << DOR0) | (1 << UPE0))) return -1;
  return r;
#endif
}

void uart0_Read(const int8_t *Data, int16_t Count) {

   while (1) {
     if (Count <= 0) break;
     *Data++ = uart0_Get();
     Count--;
   }
}

#ifdef UART0_ISR
#if defined(SIG_USART0_RECV)
SIGNAL(SIG_USART0_RECV)
#elif defined(SIG_UART0_RECV)
SIGNAL(SIG_UART0_RECV)
#elif defined(SIG_UART_RECV)
SIGNAL(SIG_UART_RECV)
#else
  #error no UART ISR vector
#endif
{
  uint8_t i = UCSR0A & ((1 << FE0) | (1 << DOR0) | (1 << UPE0));
  if (i) {
    UDR0;
// todo: error handling
  } else {
    i = uart0_rx_head;
    uart0_rx_buffer[i++] = UDR0;
    i &= UART0_RX_BUFFER_SIZE -1;
    if (i != uart0_rx_tail) {
      uart0_rx_head = i;
    } else {
      UCSR0B &= ~(1 << RXCIE0);
// todo: error handling
    }
  }
}

#if defined(SIG_USART0_DATA)
SIGNAL(SIG_USART0_DATA)
#elif defined(SIG_UART0_DATA)
SIGNAL(SIG_UART0_DATA)
#elif defined(SIG_UART_DATA)
SIGNAL(SIG_UART_DATA)
#else
  #error no UART ISR vector
#endif
{
  uint8_t i = uart0_tx_tail;
  if (i != uart0_tx_head)
  {
    UDR0 = uart0_tx_buffer[i++];
    i &= UART0_TX_BUFFER_SIZE -1;
    uart0_tx_tail = i;
  } else {
    UCSR0B &= ~(1 << UDRIE0);
  }
}
#endif
#endif


#ifdef UART1
void uart1_Init(uint16_t baud) {

  UCSR1B = 0;
  UCSR1A = 0;
  UDR1 = 0;
#if defined(U2X1)
  if (baud & (1 << 15)) UCSR1A = (1 << U2X1);
#endif
  UBRR1L = baud & 0xFF;
#if defined(UBBR1H)
  UBRR1H = (baud >> 8) & 0x0F;
#endif
#if defined(UCSR1C)
  UCSR1C = (1 << UCSZ11) | (1 << UCSZ10) | (1 << UPM11) | (1 << UPM10);
#endif
#ifdef UART1_ISR
  uart1_tx_head = 0;
  uart1_tx_tail = 0;
  uart1_rx_head = 0;
  uart1_rx_tail = 0;
  UCSR1B = (1 << TXEN1) | (1 << RXEN1) | (1 << RXCIE1);
#else
  UCSR1B = (1 << TXEN1) | (1 << RXEN1);
#endif
}

int16_t uart1_Put(uint8_t value) {

#ifdef UART1_ISR
  uint8_t h = uart1_tx_head;
  uint8_t i = h + 1;
  i &= UART1_TX_BUFFER_SIZE -1;
  while (1) {
    if (i != uart1_tx_tail) break;
    wdt_reset();
  }
  uart1_tx_buffer[h] = value;
  uart1_tx_head = i;
  UCSR1B |= (1 << UDRIE1);
#else
  while (1) {
    if (UCSR1A & (1 << UDRE1)) break;
    wdt_reset();
  }
  UDR1 = value;
#endif
  return value;
}

void uart1_Write(const int8_t *Data, int16_t Count) {

   while (1) {
     if (Count <= 0) break;
     uart1_Put(*Data++);
     Count--;
   }
}

int16_t uart1_TxSpace(void) {

#ifdef UART1_ISR
  int16_t i = uart1_tx_tail - uart1_tx_head;
  if (i <= 0) i += UART1_TX_BUFFER_SIZE;
  return i - 1;
#else
  return 255;
#endif
}

void uart1_TxFlush(void) {

#ifdef UART1_ISR
  UCSR1B &= ~(1 << UDRIE1);
  uart1_tx_head = 0;
  uart1_tx_tail = 0;
  UCSR1B |= (1 << UDRIE1);
#endif
}

void uart1_RxClear(void) {

#ifdef UART1_ISR
  UCSR1B &= ~(1 << RXCIE1);
  uart1_rx_tail = 0;
  uart1_rx_head = 0;
// todo: error handling reseting
  UCSR1B |= (1 << RXCIE1);
#endif
}

int16_t uart1_RxCount(void) {

#ifdef UART1_ISR
  int16_t i = uart1_rx_head - uart1_rx_tail;
  if (i < 0) i += UART1_RX_BUFFER_SIZE;
  return i;
#else
  return ((UCSR1A & (1 << RXC1)) ? 1 : 0);
#endif
}

int16_t uart1_Get(void) {

#ifdef UART1_ISR
  uint8_t i = uart1_rx_tail;
  while (1) {
    if (i != uart1_rx_head) break;
    wdt_reset();
  }
  uint8_t r = uart1_rx_buffer[i++];
  i &= UART1_RX_BUFFER_SIZE -1;
  uart1_rx_tail = i;
  UCSR1B |= (1 << RXCIE1);
  return r;
#else
  while (1) {
    if (UCSR1A & (1 << RXC1)) break;
    wdt_reset();
  }
  uint8_t r = UDR1;
  if (UCSR1A & ((1 << FE1) | (1 << DOR1) | (1 << UPE1))) return -1;
  return r;
#endif
}

void uart1_Read(const int8_t *Data, int16_t Count) {

   while (1) {
     if (Count <= 0) break;
     *Data++ = uart1_Get();
     Count--;
   }
}

#ifdef UART1_ISR
#if defined(SIG_USART1_RECV)
SIGNAL(SIG_USART1_RECV)
#elif defined(SIG_UART1_RECV)
SIGNAL(SIG_UART1_RECV)
#else
  #error no UART ISR vector
#endif
{
  uint8_t i = UCSR1A & ((1 << FE1) | (1 << DOR1) | (1 << UPE1));
  if (i) {
    UDR1;
// todo: error handling    
  } else {
    i = uart1_rx_head;
    uart1_rx_buffer[i++] = UDR1;
    i &= UART1_RX_BUFFER_SIZE -1;
    if (i != uart1_rx_tail) {
      uart1_rx_head = i;
    } else {
      UCSR1B &= ~(1 << RXCIE1);
// todo: error handling    
    }
  }
}

#if defined(SIG_USART1_DATA)
SIGNAL(SIG_USART1_DATA)
#elif defined(SIG_UART1_DATA)
SIGNAL(SIG_UART1_DATA)
#else
  #error no UART ISR vector
#endif
{
  uint8_t i = uart1_tx_tail;
  if (i != uart1_tx_head)
  {
    UDR1 = uart1_tx_buffer[i++];
    i &= UART1_TX_BUFFER_SIZE -1;
    uart1_tx_tail = i;
  } else {
    UCSR1B &= ~(1 << UDRIE1);
  }
}
#endif
#endif

