#ifndef INC_UART_H
#define INC_UART_H

#include <stddef.h>
#include <inttypes.h>


// #define UART0
#define UART1

#if !defined(UCSR1A)
  #undef UART1
#endif


#ifdef UART0
  #define UART0_RX_BUFFER_SIZE  16   //2,4,8,16,32,64,128,256
  #define UART0_TX_BUFFER_SIZE  16
  #define UART0_ISR
#endif

#ifdef UART1
  #define UART1_RX_BUFFER_SIZE  256
  #define UART1_TX_BUFFER_SIZE  256
  #define UART1_ISR
#endif


#ifdef XTAL
  #define UART_XTAL  XTAL
#else
  #define UART_XTAL 16000000
#endif


#define UART_BAUDRATE(baud) ((uint16_t)((UART_XTAL / (16.0 * (baud))) + 0.5) - 1)

#define B115200 (UART_BAUDRATE(115200UL))
#define B76800  (UART_BAUDRATE( 76800UL))
#define B57600  (UART_BAUDRATE( 57600UL))
#define B38400  (UART_BAUDRATE( 38400UL))
#define B28800  (UART_BAUDRATE( 28800UL))
#define B19200  (UART_BAUDRATE( 19200UL))
#define B14400  (UART_BAUDRATE( 14400UL))
#define B9600   (UART_BAUDRATE(  9600UL))
#define B4800   (UART_BAUDRATE(  4800UL))
#define B2400   (UART_BAUDRATE(  2400UL))

#if defined(U2X0) || defined(U2X1) || defined(U2X)
  #define UART_2x_BAUDRATE(baud) (((uint16_t)((UART_XTAL / (8.0 * (baud))) + 0.5) - 1) | (1 << 15))

  #define B2x115200 (UART_2x_BAUDRATE(115200UL))
  #define B2x76800  (UART_2x_BAUDRATE( 76800UL))
  #define B2x57600  (UART_2x_BAUDRATE( 57600UL))
  #define B2x38400  (UART_2x_BAUDRATE( 38400UL))
  #define B2x28800  (UART_2x_BAUDRATE( 28800UL))
  #define B2x19200  (UART_2x_BAUDRATE( 19200UL))
  #define B2x14400  (UART_2x_BAUDRATE( 14400UL))
  #define B2x9600   (UART_2x_BAUDRATE(  9600UL))
  #define B2x4800   (UART_2x_BAUDRATE(  4800UL))
  #define B2x2400   (UART_2x_BAUDRATE(  2400UL))
#endif


#ifdef UART0
void    uart0_Init(uint16_t baud);
int16_t uart0_Put(uint8_t value);
int16_t uart0_Get(void);
void    uart0_Write(const int8_t *Data, int16_t Count);
void    uart0_Read(const int8_t *Data, int16_t Count);
int16_t uart0_TxSpace(void);
void    uart0_TxFlush(void);
void    uart0_RxClear(void);
int16_t uart0_RxCount(void);
#endif

#ifdef UART1
void    uart1_Init(uint16_t baud);
int16_t uart1_Put(uint8_t value);
int16_t uart1_Get(void);
void    uart1_Write(const int8_t *Data, int16_t Count);
void    uart1_Read(const int8_t *Data, int16_t Count);
int16_t uart1_TxSpace(void);
void    uart1_TxFlush(void);
void    uart1_RxClear(void);
int16_t uart1_RxCount(void);
#endif

#endif
